# cody.py (Explorer Workshop)
# Copyright: 2024 Bradan Lane STUDIO
#
# Open Source License: MIT

'''
This CircuitPython program is a sample from the 'Hack the Badge' workshop

Implementation Notes
--------------------
**Hardware:**
    * one of the Explorer Badge designs

**Software and Dependencies:**
    * CircuitPython firmware (UF2): https://circuitpython.org
    * Adafruit CircuitPython Libraries: https://circuitpython.org/libraries

**Open Source Repository:**
    * GitLab: https://gitlab.com/bradanlane/exp_tutorial
    
'''

import time
import board
import neopixel
import digitalio


''' NOTE: The following Class has been copied from the explorer.py library '''

class expAccelerometer:
    '''
        The badge has the MXC4005XC 3-axis accelerometer.
        It has 12bit values for each axis and defaults to +/- 2G.
        The accelerometer provides 'acceleration' data for X, Y, and Z.
        It does not provide 'position' data such as angles.
        The accelerometer has an interrupt pin to quickly indicate orientation changes and shaking.

        The accelerometer is controlled using I2C.
            - I2C is a 2-wire serial interface which allows many devices to share the same wires.
            - Each I2C device, sharing the I2C wires, has its own 1-byte address.
            - CircuitPython provides a means to scan I2C for all attached devices.

        The accelerometer has an I2C address of `0x15` and is accessed using `board.I2C()`.
    '''
    __DEFAULT_ADDR        = 0x15
    __SETTING_SHAKE       = 0b00001111 # shake -Y +Y -X +X
    __SETTING_ORIENT      = 0b11000000 # orientation Z and XY
    __INTERRUPT_CHANGE    = 0x00
    __INTERRUPT_READY     = 0x01
    __READ_STATUS         = 0x02
    __STATUS_READY        = 0b00010000
    __READ_X              = 0x03
    __READ_Y              = 0x05
    __READ_Z              = 0x07
    __READ_T              = 0x09
    __SET_SHAKE_INT       = 0x0A
    __SET_TILT_INT        = 0x0B
    __READ_SETTINGS       = 0x0D
    __WRITE_SETTINGS      = 0x0D
    # settings are a bit field
    __SETTING_POWERDOWN   = 0b00000001
    __SETTING_POWERUP     = 0b00000000
    __SETTING_RANGE       = 0b01100000
    __SETTING_2G          = 0b00000000
    __SETTING_4G          = 0b00100000
    __SETTING_8G          = 0b01000000

    def __init__(self, reserve_interrupt:bool=False):
        ''' Initialize the accelerometer '''
        time.sleep(0.5)
        self.__i2c = board.I2C()
        '''
        while not self.__i2c.try_lock():
            pass
        '''
        self.__addr = expAccelerometer.__DEFAULT_ADDR
        self.__range = 0
        self.__power = 0
        self.__data = bytearray(2)
        self.__x = 0    # x axis (raw)
        self.__y = 0    # y axis (raw)
        self.__z = 0    # z axis (raw)
        self.__t = 0    # temperature (celsius)
        time.sleep(0.2) # this may not be 100% necessary but we give the device some time to settle
        self.__i2c.unlock()
        val = self.__read_byte(expAccelerometer.__READ_SETTINGS)
        self.__power = val & expAccelerometer.__SETTING_POWERDOWN
        self.__range = (0x01 << (((val & expAccelerometer.__SETTING_RANGE) >> 5) + 1)) # 2, 4, or 8
        self.update()   # get initial data to 'prime the pump'
        #print ("Data: X:{:+5d}   Y:{:+5d}   Z:{:+5d}   T:{:2d}".format(gyro.x, gyro.y, gyro.z, gyro.temperature))
        if reserve_interrupt:
            self.__interrupt = None
        else:
            self.__interrupt = digitalio.DigitalInOut(board.GP28)
            self.__interrupt.direction = digitalio.Direction.INPUT
            self.__interrupt.pull = digitalio.Pull.UP
        val = self.__write_byte(expAccelerometer.__SET_SHAKE_INT, (expAccelerometer.__SETTING_SHAKE)) # option:  | expAccelerometer.__SETTING_ORIENT
        # there are several things which can be detected; we simplify the API by generalizing to just one action
        self.__orientation = 0
        self.__shake = 0

    def __repr__(self):
        return "MXC4005XC: address:0x{:02X}: range:+/-{:d}g power:{} x:{:+05d} y:{:+05d} z:{:+05d} temperature:{:02d}".format(self.__addr, self.__range, ("DN" if self.__power else "UP"), self.__x, self.__y, self.__z, self.__t)

    def __write_byte(self, cmd, val):
        ''' internal method to write a command and a byte '''
        count = 0
        try:
            # print ("I2C Write: %02X %02X" % (cmd, val))
            if self.__i2c.try_lock():
                # command and data must be sent together for proper timing
                self.__i2c.writeto(self.__addr, bytes([cmd, val]))
                self.__i2c.unlock()
            else:
                self.__i2c.unlock()
                print ("********  I2C Write lock error  ********")
        except: # OSError as e:
            print ("********  I2C Write error  ********")
            self.__i2c.unlock()
        return val


    def __read_byte(self, cmd):
        ''' internal method to write a command and then read a byte '''
        val = 0
        try:
            # print ("I2C Read: %02X" % (cmd))
            if self.__i2c.try_lock():
                self.__i2c.writeto(self.__addr, bytes([cmd]))
                self.__i2c.readfrom_into(self.__addr, self.__data)
                val = self.__data[0]
                self.__i2c.unlock()
            else:
                self.__i2c.unlock()
                print ("********  I2C Read lock error  ********")
        except: # OSError as e:
            print ("********  I2C Read error  ********")
            self.__i2c.unlock()
        return val

    def __read_doublebyte(self, cmd):
        ''' internal method to write a command and read two successive bytes; assumes only high 12 bits are significant '''
        val = self.__read_byte(cmd)                 # read MSB
        val = (val << 8) + self.__read_byte(cmd+1)  # add MSB and LSB
        val = val // 16                             # shift out the empty 4 bits, preserving the sign
        if val > 2047:
            val = val - 4096                        # adjust to 2's compliment
        return val

    def update(self):
        ''' update should be called within the main loop to compute the read the current accelerometer data '''
        self.__x = self.__read_doublebyte(expAccelerometer.__READ_X)
        self.__y = self.__read_doublebyte(expAccelerometer.__READ_Y)
        self.__z = self.__read_doublebyte(expAccelerometer.__READ_Z)

        self.__t = 0
        self.__t = self.__read_byte(expAccelerometer.__READ_T)
        if self.__t > 127:
            self.__t = self.__t - 256
        self.__t += 25  # the base temperature reading is 25C

        RANGE = 1024    # the datasheet implies this should be 2048
        self.__p = self.__x / (RANGE * (-90.0))
        self.__r = self.__y / (RANGE * (-90.0))
        if (self.__z < 0):
            if (self.__y < 0):
                self.__r = 180.0 - self.__r
            else:
                self.__r = -180.0 - self.__r
        # print ("Data: X:%+5d    Y:%+5d    Z:%+5d    P:%+5.1f    R:%+5.1f    T:%2d" % (self.__x, self.__y, self.__z, self.__p, self.__r, self.__t))


    @property
    def interrupt(self) -> bool:
        if self.__interrupt is None:
            # clear the interrupt, regardless of the pin being controlled or not
            val = self.__read_byte(expAccelerometer.__INTERRUPT_CHANGE) # read the data
            self.__write_byte(expAccelerometer.__INTERRUPT_CHANGE, val) # clear the data
            return val != 0
        ''' return a boolean (True) if the interrupt has been triggered '''
        if self.__interrupt.value == 0: # pin is pulled LOW to indicate the interrupt has occurred
            self.__shake = 0
            self.__orientation = 0
            
            val = self.__read_byte(expAccelerometer.__INTERRUPT_CHANGE) # read the data
            self.__write_byte(expAccelerometer.__INTERRUPT_CHANGE, val) # clear the data

            #print("interrupt: shake=0x{:02X} tilt=0x{:02X}".format((val & expAccelerometer.__SETTING_SHAKE), (val & expAccelerometer.__SETTING_ORIENT) >> 6))
            self.__shake = val & expAccelerometer.__SETTING_SHAKE
            self.__orientation = val >> 6
            return True
        self.__shake = 0
        self.__orientation = 0
        return False

    @property
    def shake(self) -> bool:
        ''' return a boolean (True) if 'shake' has been detected (use after checking 'interrupt') '''
        shake = self.__shake
        self.__shake = 0
        return shake

    @property
    def orientation(self) -> int:
        ''' return the axis of the orientation change (use after checking 'interrupt') '''
        orientation = self.__orientation
        self.__orientation = 0
        return orientation

    @property
    def x(self) -> int:
        ''' return the current X axis acceleration value (int) '''
        # the datasheet says this is +/- 2048 but I'm only seeing +/- 1024
        return self.__x

    @property
    def y(self) -> int:
        ''' return the current Y axis acceleration value (int) '''
        # the datasheet says this is +/- 2048 but I'm only seeing +/- 1024
        return self.__y

    @property
    def z(self) -> int:
        ''' return the current Z axis acceleration value (int) '''
        return self.__z

    @property
    def temperature(self) -> int:
        ''' return the current celsius temperature value (int) '''
        return self.__t

    @property
    def range(self) -> int:
        ''' return the current acceleration g-force range (int) : 2, 4, or 8 '''
        val = self.__read_byte(expAccelerometer.__READ_SETTINGS)
        power = val & expAccelerometer.__SETTING_POWERDOWN
        # return 2, 4, or 8
        return (0x01 << (((val & expAccelerometer.__SETTING_RANGE) >> 5) + 1))
    @range.setter
    def range(self, range):
        ''' change the acceleration g-force range (int) : 2, 4, or 8 '''
        if range <= 2:
            range = 0
        elif range <= 4:
            range = 1
        else:
            range = 2
        range = range << 5
        self.__write_byte(expAccelerometer.__WRITE_SETTINGS, range)

    @property
    def power(self) -> bool:
        ''' return boolean (True) when the accelerometer is powered up and active '''
        val = self.__read_byte(expAccelerometer.__READ_SETTINGS)
        return (val & expAccelerometer.__SETTING_POWERDOWN)
    @power.setter
    def power(self, value:int):
        ''' set the accelerometer powered state (1=on 0=off) '''
        if value:
            ''' power up the chip after it hs been powered down '''
            self.__write_byte(expAccelerometer.__WRITE_SETTINGS, expAccelerometer.__SETTING_POWERUP)
        else:
            ''' power down the chip '''
            self.__write_byte(expAccelerometer.__WRITE_SETTINGS, expAccelerometer.__SETTING_POWERDOWN)



''' setup the Neopixel LEDs '''
NEOPIXEL_COUNT: int = 9
leds = neopixel.NeoPixel(board.NEOPIXEL, NEOPIXEL_COUNT, brightness=0.15, auto_write=True)
RED: int    = 0xFF0000
GREEN: int  = 0x00FF00
BLUE: int   = 0x0000FF

gyro = expAccelerometer()

counter = 0

try:
    while True:
        gyro.update()
        if gyro.x < 0:
            leds[0] = RED
        else:
            leds[0] = GREEN
        if gyro.y < 0:
            leds[1] = RED
        else:
            leds[1] = GREEN
        if gyro.z < 0:
            leds[2] = RED
        else:
            leds[2] = GREEN

        if gyro.interrupt:
            action = ("SHAKE" if gyro.shake else ("MOTION" if gyro.orientation else ""))
            print(counter, action)
            counter += 1

        time.sleep(0.1)
finally:
    leds.fill(0x000000)
